// Copyright Contributors to the OpenVDB Project
// SPDX-License-Identifier: MPL-2.0

// For Windows, we need these includes to ensure all OPENVDB_API
// functions/classes are compiled into the shared library.
#include "openvdb.h"
#include "Exceptions.h"
